@@ -24,17 +24,18 @@
 #include <unistd.h>		/* POSIX.1: fork */
 #include <netdb.h>		/* POSIX.1: gethostbyname, h_errno, ?hstrerror? */
 #include <arpa/inet.h>		/* BSD 4.3: inet_aton */
-#include <dirent.h>
+#include <dirent.h>		/* ? */
 #include <inttypes.h>		/* POSIX.1 (CX): PRI* */
 #include "xvasprintf.h"		/* Gnulib */
 #include "xstrndup.h"		/* Gnulib */
 #include "xalloc.h"		/* Gnulib */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #include "quotearg.h"		/* Gnulib */
+#include "fnmatch.h"		/* Gnulib/GNU Libc */
 #include "quote.h"		/* Gnulib */
+#include "stpcpy.h"		/* Gnulib/GNU Libc */
 #define _(s) gettext(s)
 #define N_(s) gettext_noop(s)
-#include "stpcpy.h"		/* Gnulib */
 #include "common/tmap.h"
 #include "common/error.h"
 #include "common/strleftcmp.h"
@@ -75,21 +76,29 @@ static void cmd_unsearch(int argc, char **argv);
 static void cmd_alias(int argc, char **argv);
 static void cmd_unalias(int argc, char **argv);
 static void cmd_shell(int argc, char **argv);
-static void shell_command_completion_selector(const char *line,int ws,int we,char *word,PtrV *results);
-static void executable_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
-static void alias_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
-static void command_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
-static void alias_command_completion_selector(const char *line,int ws,int we,char *word,PtrV *results);
+static void shell_command_completion_selector(DCCompletionInfo *ci);
+static void executable_completion_generator(DCCompletionInfo *ci);
+static void alias_completion_generator(DCCompletionInfo *ci);
+static void command_completion_generator(DCCompletionInfo *ci);
+static void alias_command_completion_selector(DCCompletionInfo *ci);
 
 typedef enum {
     DC_CMD_BUILTIN,
     DC_CMD_ALIAS,
 } DCCommandType;
 
+typedef struct _DCGetData DCGetData;
 typedef struct _DCCommand DCCommand;
 typedef struct _DCAliasCommand DCAliasCommand;
 typedef struct _DCBuiltinCommand DCBuiltinCommand;
 
+struct _DCGetData {
+    DCUserInfo *ui;
+    char *base_path;
+    uint64_t byte_count;
+    uint32_t file_count;
+};
+
 struct _DCCommand {
     char *name;
     DCCommandType type;
@@ -277,56 +286,72 @@ command_init(void)
         _("unalias NAME ..."),
         _("Remove aliases.\n"));
     add_builtin_command("shell", cmd_shell, shell_command_completion_selector,
-        _("shell COMMAND [ARGUMENTS...]"),
-        _("Execute a system command. microdc will continue in the background while "
-          "the command is executing.\n"));
+        _("shell [COMMAND [ARGUMENTS...]]"),
+        _("Execute a system command. If no arguments are specified, the current "
+          "shell will be started (SHELL environment variable or `/bin/sh' if that "
+          "is not set). microdc will continue in the background while the command "
+          "is executing.\n"));
 
     add_alias_command("ll", "ls -l");
 }
 
 void
-default_completion_selector(const char *line,int ws,int we,char *word,PtrV *results)
+default_completion_selector(DCCompletionInfo *ci)
 {
+    DCCompletionInfo ci2;
     const char *newline;
 
-    if (find_unquoted_leading_char(line, line+we, '#') != NULL)
+    if (find_unquoted_leading_char(ci->line, ci->line + ci->we, '#') != NULL)
 	return; /* completing in commented out text */
 
-    newline = find_last_unquoted_char(line, line+we, ';');
+    newline = find_last_unquoted_char(ci->line, ci->line + ci->we, ';');
     if (newline != NULL) {
-	newline = find_word_start(newline+1, line+we);
-	ws -= newline-line;
-	we -= newline-line;
-	line = newline;
+	ci2.line = find_word_start(newline+1, ci->line + ci->we);
+	ci2.ws = ci->ws - (ci2.line - ci->line);
+	ci2.we = ci->we - (ci2.line - ci->line);
+	ci2.results = ci->results;
+	fill_completion_info(&ci2);
+    } else {
+        ci2 = *ci;
     }
-    if (get_word_index(line, ws) == 0) {
-        command_completion_generator(line,ws,we,word,results);
+
+    if (ci2.word_index == 0) {
+        command_completion_generator(&ci2);
     } else {
         DCCommand *cmd;
         char *name;
 
-        name = get_word_dequoted(line, 0);
+        name = get_word_dequoted(ci2.line, 0);
         cmd = tmap_get(commands, name);
         free(name);
         if (cmd != NULL) {
             if (cmd->type == DC_CMD_BUILTIN) {
                 DCBuiltinCommand *builtin = (DCBuiltinCommand *) cmd;
 		if (builtin->completor != NULL)
-		    builtin->completor(line,ws,we,word,results);
+		    builtin->completor(&ci2);
             } else if (cmd->type == DC_CMD_ALIAS) {
                 DCAliasCommand *alias = (DCAliasCommand *) cmd;
                 int sizemod;
                 char *newline;
 
+                newline = xasprintf("%s%s", alias->alias_spec, find_word_end(ci2.line, NULL));
+                ci2.line = newline;
                 sizemod = strlen(alias->alias_spec) - strlen(cmd->name);
-                newline = xasprintf("%s%s", alias->alias_spec, find_word_end(line, NULL));
-                default_completion_selector(newline,ws+sizemod,we+sizemod,word,results);
+                ci2.ws += sizemod;
+                ci2.we += sizemod;
+                default_completion_selector(&ci2);
                 free(newline);
             }
         }
     }
+
+    if (newline != NULL) {
+        free(ci2.word_full);
+        free(ci2.word);
+    }
 }
 
+#if 0
 static char *
 replace_substr(const char *str, int sp, int ep, const char *repl)
 {
@@ -343,11 +368,13 @@ replace_substr(const char *str, int sp, int ep, const char *repl)
 
     return out;
 }
-
+#endif
 
 static void
-alias_command_completion_selector(const char *line,int ws,int we,char *word,PtrV *results)
+alias_command_completion_selector(DCCompletionInfo *ci)
 {
+/* XXX not implemented yet */
+#if 0
     char *newline;
     char *newword;
     int newwe;
@@ -403,30 +430,31 @@ alias_command_completion_selector(const char *line,int ws,int we,char *word,PtrV
     }
 
     free(newword);
+#endif
 }
 
 static void
-alias_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
+alias_completion_generator(DCCompletionInfo *ci)
 {
     TMapIterator it;
 
-    tmap_iterator_partial(commands, &it, word, (comparison_fn_t) strleftcmp);
+    tmap_iterator_partial(commands, &it, ci->word, (comparison_fn_t) strleftcmp);
     while (it.has_next(&it)) {
         DCCommand *cmd = it.next(&it);
         if (cmd->type == DC_CMD_ALIAS)
-            ptrv_append(results, new_completion_entry(cmd->name, NULL));
+            ptrv_append(ci->results, new_completion_entry(cmd->name, NULL));
     }
 }
 
 static void
-command_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
+command_completion_generator(DCCompletionInfo *ci)
 {
     TMapIterator it;
 
-    tmap_iterator_partial(commands, &it, word, (comparison_fn_t) strleftcmp);
+    tmap_iterator_partial(commands, &it, ci->word, (comparison_fn_t) strleftcmp);
     while (it.has_next(&it)) {
         DCCommand *cmd = it.next(&it);
-        ptrv_append(results, new_completion_entry(cmd->name, NULL));
+        ptrv_append(ci->results, new_completion_entry(cmd->name, NULL));
     }
 }
 
@@ -503,7 +531,7 @@ command_execute(const char *line)
  * executable matches in other dirs in PATH.
  */
 static void
-executable_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
+executable_completion_generator(DCCompletionInfo *ci)
 {
     char *path;
     char *p1;
@@ -512,7 +540,7 @@ executable_completion_generator(const char *line,int ws,int we,char *word,PtrV *
     char *dir_part;
     bool path_has_cwd = false;
 
-    get_file_dir_part(word, &dir_part, &file_part);
+    get_file_dir_part(ci->word, &dir_part, &file_part);
 
     path = getenv("PATH");
     if (path == NULL) {
@@ -532,13 +560,13 @@ executable_completion_generator(const char *line,int ws,int we,char *word,PtrV *
         if (dh != NULL) {
             struct dirent *ent;
             while ((ent = readdir(dh)) != NULL) { /* ignore errors */
-                if (strleftcmp(word, ent->d_name) == 0) {
+                if (strleftcmp(ci->word, ent->d_name) == 0) {
                     char *full;
                     full = catfiles(p, ent->d_name);
                     if (access(full, X_OK) == 0) { /* ignore errors */
                         struct stat sb;
                         if (stat(full, &sb) >= 0 && S_ISREG(sb.st_mode))
-                            ptrv_append(results, new_completion_entry(ent->d_name, ent->d_name));
+                            ptrv_append(ci->results, new_completion_entry(ent->d_name, ent->d_name));
                     }
                     /* XXX: use is_directory instead of that stat above */
                     free(full);
@@ -549,9 +577,9 @@ executable_completion_generator(const char *line,int ws,int we,char *word,PtrV *
     }
 
     if (*dir_part != '\0' || path_has_cwd) {
-        local_fs_completion_generator(line,ws,we,word,results,DC_CPL_DIR|DC_CPL_EXE|DC_CPL_DOT);
-    } else if (results->cur == 0) {
-        local_fs_completion_generator(line,ws,we,word,results,DC_CPL_DIR|DC_CPL_DOT);
+        local_fs_completion_generator(ci,DC_CPL_DIR|DC_CPL_EXE|DC_CPL_DOT);
+    } else if (ci->results->cur == 0) {
+        local_fs_completion_generator(ci,DC_CPL_DIR|DC_CPL_DOT);
     }
 
     free(dir_part);
@@ -559,12 +587,12 @@ executable_completion_generator(const char *line,int ws,int we,char *word,PtrV *
 }
 
 static void
-shell_command_completion_selector(const char *line,int ws,int we,char *word,PtrV *results)
+shell_command_completion_selector(DCCompletionInfo *ci)
 {
-    if (get_word_index(line, ws) == 1) {
-        executable_completion_generator(line,ws,we,word,results);
+    if (ci->word_index == 1) {
+        executable_completion_generator(ci);
     } else {
-        local_path_completion_generator(line,ws,we,word,results);
+        local_path_completion_generator(ci);
     }
 }
 
@@ -652,8 +680,21 @@ cmd_shell(int argc, char **argv)
     }
     if (shell_child == 0) {
         warn_writer = default_warn_writer;
-        if (execvp(argv[1], argv+1) < 0)
-	    die(_("%s: cannot execute - %s\n"), quotearg(argv[1]), errstr); /* die OK */
+        if (argc <= 1) {
+            char *args[2] = { NULL, NULL };
+            char *shell;
+
+            shell = getenv("SHELL");
+            if (shell != NULL)
+                args[0] = shell;
+            else
+                args[0] = "/bin/sh";
+            if (execvp(args[0], args) < 0)
+	        die(_("%s: cannot execute - %s\n"), quotearg(args[0]), errstr); /* die OK */
+        } else {
+            if (execvp(argv[1], argv+1) < 0)
+	        die(_("%s: cannot execute - %s\n"), quotearg(argv[1]), errstr); /* die OK */
+        }
         exit(EXIT_FAILURE); /* shouldn't get here */
     }
 }
@@ -727,9 +768,10 @@ cmd_say(int argc, char **argv)
     }
     tmp = join_strings(argv+1, argc-1, ' ');
     msg = escape_message(tmp);
-    free(tmp);
+    /* Don't print on screen - hub will send us the message back. */
     hub_putf("<%s> %s|", my_nick, msg); /* Ignore error */
     free(msg);
+    free(tmp);
 }
 
 static void
@@ -755,16 +797,16 @@ cmd_msg(int argc, char **argv)
 
     tmp = join_strings(argv+2, argc-2, ' ');
     msg = escape_message(tmp);
-    free(tmp);
-    hub_putf("$To: %s From: %s $<%s> %s|", ui->nick, my_nick, my_nick, tmp); /* Ignore error */
+    screen_putf("Private to %s: <%s> %s\n", quotearg_n(0, ui->nick), quotearg_n(1, my_nick), tmp);
+    hub_putf("$To: %s From: %s $<%s> %s|", ui->nick, my_nick, my_nick, msg); /* Ignore error */
     free(msg);
+    free(tmp);
 }
 
 static void
 cmd_raw(int argc, char **argv)
 {
     char *msg;
-    char *tmp;
     
     if (argc <= 1) {
     	screen_putf(_("Usage: %s DATA...\n"), argv[0]);
@@ -774,9 +816,8 @@ cmd_raw(int argc, char **argv)
     	screen_putf(_("Not connected.\n"));
 	return;
     }
-    tmp = join_strings(argv+1, argc-1, ' ');
-    msg = escape_message(tmp);
-    free(tmp);
+    msg = join_strings(argv+1, argc-1, ' ');
+    screen_putf("Raw to hub: %s\n", msg);
     hub_putf("%s", msg); /* Ignore error */
     free(msg);
 }
@@ -1111,104 +1152,6 @@ cmd_ls(int argc, char **argv)
     optparser_free(p);
 }
 
-/* XXX: move to download.c/queue.c/browse.c? */
-static bool
-append_download_dir(DCUserInfo *ui, DCFileList *node, char *base_path, uint64_t *bytes, uint32_t *files)
-{
-    if (node->type == DC_TYPE_REG) {
-	DCQueuedFile *queued;
-    	char *path = filelist_get_path(node);
-
-	if (ptrv_find(ui->download_queue, path, (comparison_fn_t) queued_file_cmp) >= 0) {
-	    screen_putf(_("Queue already contains this file, ignoring\n"));
-	    free(path);
-	    return false;
-	}
-
-	queued = xmalloc(sizeof(DCQueuedFile));
-	queued->filename = path;
-	queued->base_path = xstrdup(base_path);
-	queued->flag = DC_TF_NORMAL;
-	ptrv_append(ui->download_queue, queued);
-
-	(*bytes) += node->u.reg.size;
-	(*files) ++;
-	return true;
-    }
-    if (node->type == DC_TYPE_DIR) {
-    	HMapIterator it;
-	bool s = false;
-
-	hmap_iterator(node->u.dir.children, &it);
-	while (it.has_next(&it))
-	    s = append_download_dir(ui, it.next(&it), base_path, bytes, files) || s;
-	return s;
-    }
-    return false;
-}
-
-/* XXX: move to download.c/queue.c/browse.c? */
-static void
-cmd_get(int argc, char **argv)
-{
-    uint32_t c;
-
-    /* Note: This command assumes that we are browsing some user.
-     * If we in the future would allow browse-less getting,
-     * i.e. get NICK FILE, then we must here add additional checks
-     * found in cmd_browse, such as strcmp(my_nick, nick)==0 etc.
-     */
-
-    if (argc == 1) {
-    	screen_putf(_("Usage: %s FILE ...\n"), argv[0]);
-	return;
-    }
-    if (browse_list == NULL) {
-	screen_putf(_("Not browsing any user.\n"));
-	return;
-    }
-    if (browsing_myself) {
-	screen_putf(_("Cannot download files from myself.\n"));
-	return;
-    }
-
-    for (c = 1; c < argc; c++) {
-        DCFileList *node;
-        char *path;
-        uint64_t bytes = 0;
-        uint32_t files = 0;
-
-        path = apply_cwd(argv[c]);
-        node = filelist_lookup(browse_list, path);
-        if (node == NULL) {
-            screen_putf(_("%s: No such file or directory\n"), quotearg(argv[c]));
-            free(path);
-            return;
-        }
-        free(path);
-
-        if (append_download_dir(browse_user, node, browse_path, &bytes, &files)) {
-            char *fmt1 = ngettext("byte", "bytes", bytes);
-            char *fmt2 = ngettext("file", "files", files);
-            path = filelist_get_path(node);
-            screen_putf(_("Downloading %s (%" PRId64 " %s in %d %s)\n"), quotearg(path), bytes, fmt1, files, fmt2); /* " */
-            free(path);
-        } else {
-            /* XXX: hmm, shouldn't we return even if queue is non-empty? */
-            if (browse_user->download_queue->cur == 0) {
-                screen_putf(_("%s: No files to download (empty directories).\n"), quotearg(argv[c]));
-                continue;
-            }
-        }
-    }
-
-    if (!has_user_conn(browse_user, DC_DIR_RECEIVE) && browse_user->conn_count < DC_USER_MAX_CONN) {
-        hub_connect_user(browse_user); /* Ignore errors */
-    } else {
-        screen_putf(_("No free connections. Queued files for download.\n"));
-    }
-}
-
 /* XXX: move to download.c/queue.c/browse.c? */
 static void
 cmd_queue(int argc, char **argv)
@@ -1661,3 +1604,157 @@ update_prompt(void)
 	set_screen_prompt("%s> ", PACKAGE);
     }
 }
+
+static void
+append_download_file(DCUserInfo *ui, DCFileList *node, DCFileList *basenode, uint32_t *file_count, uint64_t *byte_count)
+{
+    if (node->type == DC_TYPE_REG) {
+        DCQueuedFile *queued;
+        char *path;
+
+        if (ui != NULL) { /* if NULL then we're browsing ourselves */
+            path = filelist_get_path(node);
+            if (ptrv_find(ui->download_queue, path, (comparison_fn_t) queued_file_cmp) >= 0) {
+                screen_putf(_("Queue already contains this file, ignoring\n"));
+                free(path);
+                return;
+            }
+
+            queued = xmalloc(sizeof(DCQueuedFile));
+            queued->filename = path;
+            queued->base_path = filelist_get_path(basenode);
+            queued->flag = DC_TF_NORMAL;
+            ptrv_append(ui->download_queue, queued);
+        }
+
+        (*byte_count) += node->u.reg.size;
+        (*file_count)++;
+    }
+    else if (node->type == DC_TYPE_DIR) {
+        HMapIterator it;
+
+        hmap_iterator(node->u.dir.children, &it);
+        while (it.has_next(&it))
+            append_download_file(ui, it.next(&it), basenode, file_count, byte_count);
+    }
+}
+
+#if 0
+static bool
+has_wildcards(const char *str)
+{
+    return strpbrk(str, "*?") != NULL;
+}
+
+/* This function will strip leading slashes in WILD.
+ * It assumes that WILD is relative to NODE.
+ */
+static void
+wildcard_get(DCFileList *node, char *wild, DCGetData *gd)
+{
+    HMapIterator it;
+    char *next;
+
+    for (; *wild == '/'; wild++);
+    next = strchr(wild, '/');
+
+    if (next != NULL) {
+        *next = '\0';
+        if (has_wildcards(wild)) {
+            hmap_iterator(node->u.dir.children, &it);
+            while (it.has_next(&it)) {
+                node = it.next(&it);
+                /* At this point we can only match directories. */
+                if (node->type == DC_TYPE_DIR) {
+                    if (fnmatch(wild, node->name, 0) == 0)
+                        wildcard_get(node, next+1, gd);
+                }
+            }
+        } else {
+            node = hmap_get(node->u.dir.children, wild);
+            if (node != NULL)
+                wildcard_get(node, next+1, gd);
+        }
+        *next = '/';
+        return;
+    }
+
+    if (has_wildcards(wild)) {
+        hmap_iterator(node->u.dir.children, &it);
+        while (it.has_next(&it)) {
+            node = it.next(&it);
+            if (fnmatch(wild, node->name, 0) == 0)
+                append_download_file(node, gd);
+        }
+    } else {
+        node = hmap_get(node->u.dir.children, wild);
+        if (node != NULL)
+            append_download_file(node, gd);
+    }
+}
+#endif
+
+/* Note: This command assumes that we are browsing some user.
+ * If we in the future would allow browse-less getting,
+ * i.e. get NICK FILE, then we must here add additional checks
+ * found in cmd_browse, such as strcmp(my_nick, nick)==0 etc.
+ */
+static void
+cmd_get(int argc, char **argv)
+{
+    uint32_t c;
+    bool dl_some = false;
+
+    if (argc == 1) {
+        screen_putf(_("Usage: %s FILE ...\n"), argv[0]);
+        return;
+    }
+    if (browse_list == NULL) {
+        screen_putf(_("Not browsing any user.\n"));
+        return;
+    }
+    /*if (browsing_myself) {
+        screen_putf(_("Cannot download files from myself.\n"));
+        return;
+    }*/
+
+    for (c = 1; c < argc; c++) {
+        DCFileList *basenode;
+        bool quoted = false;
+        uint64_t byte_count = 0;
+        uint32_t file_count = 0;
+        PtrV *results;
+        int d;
+
+        results = ptrv_new();
+        if (has_leading_slash(argv[c])) {
+            basenode = browse_list;
+        } else {
+            basenode = filelist_lookup(browse_list, browse_path);
+        }
+        remote_wildcard_expand(argv[c], &quoted, basenode, results);
+        for (d = 0; d < results->cur; d++) {
+            DCFileList *node = results->buf[d];
+            screen_putf(_("Matched %s\n"), quotearg(node->name));
+            append_download_file(browse_user, node, basenode, &file_count, &byte_count);
+        }
+        ptrv_free(results);
+
+        if (file_count > 0) {
+            screen_putf(_("Downloading %" PRId64 " %s in %" PRId32 " %s\n"),
+                byte_count, ngettext("byte", "bytes", byte_count),
+                file_count, ngettext("file", "files", file_count));
+            dl_some = true;
+        } else {
+            screen_putf(_("%s: No files to download.\n"), quotearg(argv[c]));
+        }
+    }
+
+    if (dl_some && !browsing_myself) {
+        if (!has_user_conn(browse_user, DC_DIR_RECEIVE) && browse_user->conn_count < DC_USER_MAX_CONN) {
+            hub_connect_user(browse_user); /* Ignore errors */
+        } else {
+            screen_putf(_("No free connections. Queued files for download.\n"));
+        }
+    }
+}
