@@ -1,6 +1,6 @@
 /* connection.c - Functions generic to hub and user connections
  *
- * Copyright (C) 2004-2005 Oskar Liljeblad
+ * Copyright (C) 2004, 2005 Oskar Liljeblad
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -12,9 +12,9 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU Library General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation,
+ * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #include <config.h>
@@ -32,11 +32,11 @@
  *
  * XXX: this should be generalized and the error message removed.
  */
-uint8_t *
-decode_lock(const uint8_t *lock, int locklen, uint32_t basekey)
+char *
+decode_lock(const char *lock, size_t locklen, uint32_t basekey)
 {
     uint8_t key[locklen];
-    uint8_t *outkey;
+    char *outkey;
     int c;
     int d;
 
@@ -126,7 +126,7 @@ escape_message(char *str)
 /* Dump a received or sent piece of data.
  */
 void
-dump_command(const char *header, const char *buf, int len)
+dump_command(const char *header, const char *buf, size_t len)/* XXX: uint32_t=>size_t, same for byteq */
 {
     /* The following test is only to speed up things... */
     if (display_flags & DC_DF_DEBUG || log_flags & DC_DF_DEBUG) {
