@@ -29,6 +29,7 @@
 #include "xstrndup.h"		/* Gnulib */
 #include "xalloc.h"		/* Gnulib */
 #include "quotearg.h"		/* Gnulib */
+#include "memmem.h"		/* Gnulib */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
 #define N_(s) gettext_noop(s)
@@ -44,7 +45,7 @@ char *hub_name = NULL;
 bool
 send_my_info(void)
 {
-    return hub_putf("$MyINFO $ALL %s %s<%s V:%s,M:%c,H:1/0/0,S:%d>$ $%s%c$%s$%" PRId64 "$|",
+    return hub_putf("$MyINFO $ALL %s %s<%s V:%s,M:%c,H:1/0/0,S:%d>$ $%s%c$%s$%" PRId64 "$|", /* " */
 	  my_nick,
 	  my_description, "++", "0.401", is_active ? 'A':'P', my_ul_slots,
 	  my_speed,
@@ -54,62 +55,61 @@ send_my_info(void)
 }
 
 void
-say_user_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
+say_user_completion_generator(DCCompletionInfo *ci)
 {
     HMapIterator it;
 
+    /* XXX: perhaps hub_users should be made a tmap? to speed up things */
     hmap_iterator(hub_users, &it);
     while (it.has_next(&it)) {
     	DCUserInfo *ui = it.next(&it);
 
-	if (strleftcmp(word, ui->nick) == 0) {
+	if (strleftcmp(ci->word, ui->nick) == 0) {
 	    DCCompletionEntry *entry;
-	    entry = new_completion_entry(ui->nick, NULL);
-	    entry->input_single_format = "%s:";
-	    ptrv_append(results, entry);
+	    entry = new_completion_entry(NULL, NULL); /* XXX: use the _full version */
+	    entry->input = quote_string(ui->nick, ci->word_full[0] == '"', true);
+	    entry->display = xstrdup(ui->nick);
+	    entry->input_single_fmt = "%s: ";
+	    entry->finalize = false;
+	    entry->quoted = true; /* entry->input and _single_fmt are already quoted */
+	    ptrv_append(ci->results, entry);
 	}
     }
 }
 
 void
-user_or_myself_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
+user_or_myself_completion_generator(DCCompletionInfo *ci)
+{
+    if (strleftcmp(ci->word, my_nick) == 0)
+        ptrv_append(ci->results, new_completion_entry(my_nick, NULL));
+    user_completion_generator(ci);
+}
+
+void
+user_completion_generator(DCCompletionInfo *ci)
 {
     HMapIterator it;
 
+    /* XXX: what if we are self found in this list? conflict with user_or_myself_completion_generator */
     hmap_iterator(hub_users, &it);
-    if (strleftcmp(word, my_nick) == 0)
-        ptrv_append(results, new_completion_entry(my_nick, NULL));
-
     while (it.has_next(&it)) {
     	DCUserInfo *ui = it.next(&it);
-	if (strleftcmp(word, ui->nick) == 0)
-	    ptrv_append(results, new_completion_entry(ui->nick, NULL));
+	if (strleftcmp(ci->word, ui->nick) == 0)
+	    ptrv_append(ci->results, new_completion_entry(ui->nick, NULL));
     }
 }
 
 void
-user_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
+user_with_queue_completion_generator(DCCompletionInfo *ci)
 {
     HMapIterator it;
 
+    /* XXX: for completion speed, maintain a separate TMap for users with queue? */
     hmap_iterator(hub_users, &it);
     while (it.has_next(&it)) {
     	DCUserInfo *ui = it.next(&it);
-	if (strleftcmp(word, ui->nick) == 0)
-	    ptrv_append(results, new_completion_entry(ui->nick, NULL));
-    }
-}
-
-void
-user_with_queue_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
-{
-    HMapIterator it;
-
-    hmap_iterator(hub_users, &it);
-    while (it.has_next(&it)) {
-    	DCUserInfo *ui = it.next(&it);
-	if (ui->download_queue->cur > 0 && strleftcmp(word, ui->nick) == 0)
-	    ptrv_append(results, new_completion_entry(ui->nick, NULL));
+	if (ui->download_queue->cur > 0 && strleftcmp(ci->word, ui->nick) == 0)
+	    ptrv_append(ci->results, new_completion_entry(ui->nick, NULL));
     }
 }
 
