@@ -39,12 +39,14 @@
 #include "xalloc.h"		/* Gnulib */
 #include "quotearg.h"		/* Gnulib */
 #include "sig2str.h"		/* Gnulib */
+#include "getline.h"		/* Gnulib/GNU Libc */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
 #define N_(s) gettext_noop(s)
 #include "common/error.h"
 #include "common/strleftcmp.h"
 #include "common/tempdir.h"
+#include "common/tempfailure.h"
 #include "ipc.h"
 #include "microdc.h"
 
@@ -476,15 +478,16 @@ get_user_conn_status(DCUserConn *uc)
 }*/
 
 void
-transfer_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
+transfer_completion_generator(DCCompletionInfo *ci)
 {
     HMapIterator it;
 
+    /* XXX: perhaps use a tmap for transfers as well? to speed up this */
     hmap_iterator(user_conns, &it);
     while (it.has_next(&it)) {
     	DCUserConn *uc = it.next(&it);
-	if (strleftcmp(word, uc->name) == 0)
-	    ptrv_append(results, new_completion_entry(uc->name, NULL));
+	if (strleftcmp(ci->word, uc->name) == 0)
+	    ptrv_append(ci->results, new_completion_entry(uc->name, NULL));
     }
 }
 
@@ -586,7 +589,7 @@ run_script(const char *filename, bool allow_missing)
     	    warn(_("%s: Cannot open file - %s\n"), quotearg(filename), errstr);
     	return;
     }
-    while (getline(&line, &line_len, file) >= 0) {
+    while (getline(&line, &line_len, file) > 0) {
     	int c;
 	for (c = strlen(line)-1; c > 0 && (line[c] == '\n' || line[c] == '\r'); c--)
 	    line[c] = '\0';
