@@ -133,6 +133,7 @@ typedef struct _DCUserConn DCUserConn;
 typedef struct _DCUserInfo DCUserInfo;
 typedef struct _DCFileList DCFileList;
 typedef struct _DCCompletionEntry DCCompletionEntry;
+typedef struct _DCCompletionInfo DCCompletionInfo;
 typedef struct _DCSearchSelection DCSearchSelection;
 typedef struct _DCSearchString DCSearchString;
 typedef struct _DCSearchRequest DCSearchRequest;
@@ -141,7 +142,7 @@ typedef struct _DCSearchResponse DCSearchResponse;
 typedef struct _DCQueuedFile DCQueuedFile;
 typedef struct _DCVariable DCVariable;
 
-typedef void (*DCCompletorFunction)(const char *line,int ws,int we,char *word,PtrV *results);
+typedef void (*DCCompletorFunction)(DCCompletionInfo *ci);
 typedef void (*DCBuiltinCommandHandler)(int argc, char **argv);
 typedef void (*ScreenWriter)(DCDisplayFlag flag, const char *format, va_list args);
 
@@ -183,18 +184,54 @@ struct _DCSearchRequest {
     PtrV *responses;
 };
 
+struct _DCCompletionInfo {
+    const char *line;
+    int ws;             /* may be modified */
+    int we;             /* may not be modified. */
+    char *word;         /* dequoted word */
+    char *word_full;	/* word not touched */
+    int word_index;
+    PtrV *results;
+};
+
 struct _DCCompletionEntry {
-    char *input;
-    char *input_format;
+    /* What to display for this completion result.
+     * This should not be escaped or quoted, regardless
+     * of the value of quoted.
+     */
     char *display;
-    char *input_single_format;
-    char *input_single_full_format;
-    char input_char;
-    bool inhibit_quote; /* inhibit closing quote character? */
-//  char input_char; /* character to append to input line */
-//  char display_char; /* character to display */
-//  char input_append_single; /* append to line buffer when this entry is the only match */
-//  char input_append_single_full; /* as above, but only when the full string was entered before */
+    /* display_fmt is a printf-like format which will
+     * be passed display for first and only argument when
+     * creating the final string to display. This field
+     * is usually just "%s".
+     */
+    const char *display_fmt;
+    /* What to place in the command line for this result.
+     */
+    char *input;
+    /* input_fmt is a printf-like format which will be
+     * passed input for the first and only argument when
+     * creating the final input string.
+     */
+    const char *input_fmt;
+    /* If input_single_fmt is non-NULL, it will be
+     * used as the printf-like format if input_fmt is the
+     * only match.
+     */
+    const char *input_single_fmt;
+    /* If finalize is true, then the completion will end if this
+     * this entry is the single match. It basicly means that
+     * a space will be appended so that the caret will be moved
+     * to a new word. If the word is quoted, a closing quote
+     * will be added as well.
+     */
+    bool finalize;
+    /* quoted should be true if input is already quoted.
+     * if quoted is true and input starts with a double quote,
+     * input must also end in a double quote regardless of
+     * finalize.
+     */
+    bool quoted;
 };
 
 struct _DCFileList {
@@ -324,14 +361,14 @@ void hub_now_writable(void);
 bool send_my_info(void);
 extern struct sockaddr_in hub_addr;
 extern char *hub_name;
-void say_user_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
-void user_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
-void user_or_myself_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
-void user_with_queue_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
+void say_user_completion_generator(DCCompletionInfo *ci);
+void user_completion_generator(DCCompletionInfo *ci);
+void user_or_myself_completion_generator(DCCompletionInfo *ci);
+void user_with_queue_completion_generator(DCCompletionInfo *ci);
 
 /* command.c */
 void command_execute(const char *line);
-void default_completion_selector(const char *line,int ws,int we,char *word,PtrV *results);
+void default_completion_selector(DCCompletionInfo *ci);
 /*bool append_download_dir(DCUserInfo *ui, DCFileList *node, char *base_path, uint64_t *bytes, uint32_t *files) XXX move transfer.c? */
 void update_prompt(void);
 void command_init(void);
@@ -352,8 +389,12 @@ bool set_log_file(const char *new_filename, bool verbose);
 void sorted_list_completion_generator(const char *base, PtrV *results,
     void *items, size_t item_count, size_t item_size, size_t key_offset); /* completion */
 DCCompletionEntry *new_completion_entry(const char *input, const char *display); /* completion */
+DCCompletionEntry *new_completion_entry_full(char *input, char *display, const char *input_fmt, const char *display_fmt, bool finalize, bool quoted);
 void free_completion_entry(DCCompletionEntry *entry); /* completion */
 void get_file_dir_part(const char *word,char **dir_part,const char **file_part); /* completion */
+void fill_completion_info(DCCompletionInfo *ci);
+char *quote_string(const char *str, bool dquotes, bool finalize);
+char *filename_quote_string(const char *str, bool dquotes, bool finalize);
 
 /* huffman.c */
 char *huffman_decode(const uint8_t *data, uint32_t data_size, uint32_t *out_size);
@@ -372,7 +413,7 @@ bool set_share_dir(const char *dir);
 bool has_user_conn(DCUserInfo *info, DCTransferDirection dir);
 uint32_t get_user_conn_count(DCUserInfo *info);
 bool get_package_file(const char *name, char **outname);
-void transfer_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
+void transfer_completion_generator(DCCompletionInfo *ci);
 void user_conn_cancel(DCUserConn *uc);
 void warn_file_error(int res, bool write, const char *filename);
 void warn_socket_error(int res, bool write, const char *subject);
@@ -390,14 +431,16 @@ char *resolve_download_file(DCUserInfo *ui, const char *inname, const char *base
 bool filelist_create(const char *dir);
 char *translate_local_to_remote(const char *localname);
 char *translate_remote_to_local(const char *remotename);
-void local_fs_completion_generator(const char *line,int ws,int we,char *word,PtrV *results, DCFSCompletionFlags flags);
-void local_path_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
-void local_dir_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
-void remote_path_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
-void remote_dir_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
+void local_fs_completion_generator(DCCompletionInfo *ci, DCFSCompletionFlags flags);
+void local_path_completion_generator(DCCompletionInfo *ci);
+void local_dir_completion_generator(DCCompletionInfo *ci);
+void remote_path_completion_generator(DCCompletionInfo *ci);
+void remote_dir_completion_generator(DCCompletionInfo *ci);
 char *apply_cwd(const char *path); /* XXX: move transfer.c? */
 extern DCFileList *our_filelist;
 void filelist_list_recursively(DCFileList *node, char *basepath, int skip);
+void remote_wildcard_expand(char *mtachpath, bool *quotedptr, DCFileList *basenode, PtrV *results);
+bool has_leading_slash(const char *str);
     
 /* connection.c */
 char *decode_lock(const char *lock, size_t locklen, uint32_t basekey);
@@ -414,6 +457,7 @@ bool fd_set_status_flags(int fd, bool set, int modflags);
 char *getenv_default(const char *name, char *defvalue);
 #define IS_CURRENT_DIR(x) ((x)[0] == '.' && (x)[1] == '\0')
 #define IS_PARENT_DIR(x)  ((x)[0] == '.' && (x)[1] == '.' && (x)[2] == '\0')
+#define IS_SPECIAL_DIR(x) (IS_CURRENT_DIR(x) || IS_PARENT_DIR(x))
 #define COMPARE_RETURN(a,b) { if ((a) < (b)) return -1; if ((a) > (b)) return 1; }
 #define COMPARE_RETURN_FUNC(f) { int _c = (f); if (_c != 0) return _c; }
 char *sockaddr_in_str(struct sockaddr_in *addr);
@@ -438,6 +482,6 @@ char *search_selection_to_string(DCSearchSelection *sr);
 
 /* variables.c */
 void cmd_set(int argc, char **argv);
-void set_command_completion_selector(const char *line,int ws,int we,char *word,PtrV *results);
+void set_command_completion_selector(DCCompletionInfo *ci);
 
 #endif
