@@ -37,10 +37,12 @@
 #include "xstrndup.h"		/* Gnulib */
 #include "xvasprintf.h"		/* Gnulib */
 #include "quotearg.h"		/* Gnulib */
+#include "minmax.h"		/* Gnulib */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
 #define N_(s) gettext_noop(s)
 #include "microdc.h"
+#include "common/strbuf.h"
 #include "common/error.h"
 #include "common/strleftcmp.h"
 #include "common/bksearch.h"
@@ -157,6 +159,7 @@ fix_winch(void)
     return 0;
 }
 
+/* XXX: move this to strutil.c or something? */
 void
 get_file_dir_part(const char *word, char **dir_part, const char **file_part)
 {
@@ -174,17 +177,31 @@ get_file_dir_part(const char *word, char **dir_part, const char **file_part)
     }
 }
 
+DCCompletionEntry *
+new_completion_entry_full(char *input, char *display, const char *input_fmt, const char *display_fmt, bool finalize, bool quoted)
+{
+    DCCompletionEntry *entry = xmalloc(sizeof(DCCompletionEntry));
+    entry->input = input;
+    entry->display = display;
+    entry->input_fmt = input_fmt;
+    entry->input_single_fmt = NULL;
+    entry->display_fmt = display_fmt;
+    entry->finalize = finalize;
+    entry->quoted = quoted;
+    return entry;
+}
+
 DCCompletionEntry *
 new_completion_entry(const char *input, const char *display)
 {
     DCCompletionEntry *entry = xmalloc(sizeof(DCCompletionEntry));
     entry->input = input == NULL ? NULL : xstrdup(input);
     entry->display = display == NULL ? entry->input : xstrdup(display);
-    entry->input_char = ' ';
-    entry->input_format = "%s";
-    entry->input_single_format = NULL;
-    entry->input_single_full_format = NULL;
-    entry->inhibit_quote = false;
+    entry->display_fmt = "%s";
+    entry->input_fmt = "%s";
+    entry->input_single_fmt = NULL;
+    entry->finalize = true;
+    entry->quoted = false;
     return entry;
 }
 
@@ -197,93 +214,6 @@ free_completion_entry(DCCompletionEntry *entry)
     free(entry);
 }
 
-static char **
-attempted_completion(const char *text, int start, int end)
-{
-    char **matches = NULL;
-    char *word;
-    bool quoted;
-    PtrV *results = ptrv_new();
-
-    /* Readline is strange. When completing a quoted word,
-     * sometimes the start position is just after the
-     * first quote. And sometimes it is on the first quote.
-     * We need to manually detect which case it is, because
-     * we need the quote.
-     */
-
-    quoted = char_is_quoted(rl_line_buffer, start)
-               && ((start > 0 && rl_line_buffer[start-1] == '"') || rl_line_buffer[start] == '"');
-    if (quoted && start > 0 && rl_line_buffer[start-1] == '"')
-        start--;
-    word = dequote_words_full(rl_line_buffer+start, false, false, true, true, rl_line_buffer+end);
-
-    /* Do not attempt readline's default filename completion if our
-     * completors fails to return any results.
-     */
-    rl_attempted_completion_over = 1;
-    default_completion_selector(rl_line_buffer,start,end,word,results);
-
-//printf("%d:%d [%s:%s]\n", start,end,text,rl_line_buffer);
-    if (results->cur == 1) {
-        DCCompletionEntry *entry = results->buf[0];
-        char *match;
-
-        if (entry->input_single_full_format != NULL && strcmp(word, entry->input) == 0) {
-            match = xasprintf(entry->input_single_full_format, entry->input);
-        } else if (entry->input_single_format != NULL) {
-	    match = xasprintf(entry->input_single_format, entry->input);
-	} else {
-            match = xasprintf(entry->input_format, entry->input);
-        }
-	matches = xmalloc(sizeof(char *) * 2);
-        matches[0] = quote_word_full(match, quoted, entry->inhibit_quote, ";", "#", false, true, true, true);
-
-//printf("\n%d [[%s]=>[%s]] : %s\n", quoted, match, matches[0], word);
-        matches[1] = NULL;
-        rl_completion_append_character = entry->input_char;
-        /*rl_completion_suppress_quote = entry->inhibit_quote;*/
-        free_completion_entry(entry);
-        free(match);
-    } else if (results->cur > 1) {
-        char *match = NULL;
-        int match_len = 0;
-        int c;
-
-    	matches = xmalloc(sizeof(char *) * (results->cur + 2));
-        for (c = 0; c < results->cur; c++) {
-            DCCompletionEntry *entry = results->buf[c];
-
-            /* Determine common leading characters for all matches. */
-            if (c == 0) {
-                match = xstrdup(entry->input);
-                match_len = strlen(match);
-            } else {
-                int d;
-                char c1, c2;
-                for (d = 0; (c1 = match[d]) && (c2 = entry->input[d]); d++) {
-	            if (c1 != c2)
-	    	        break;
-                }
-                match_len = MIN(match_len, d);
-            }
-
-	    matches[c+1] = quote_word_full(entry->display, false, true, "", "", false, true, true, false);
-	    free_completion_entry(entry);
-        }
-
-        match[match_len] = '\0';
-        rl_completion_append_character = ' ';
-	matches[0] =  quote_word_full(match, quoted, false, ";", "#", false, true, true, true);
-        matches[c+1] = NULL;
-        free(match);
-    }
-
-    ptrv_free(results);
-    free(word);
-    return matches;
-}
-
 /* This function is called via the warn_writer variable by warn() and
  * die() in error.c to print messages on screen properly.
  */
@@ -443,9 +373,9 @@ screen_finish(void)
     	//rl_basic_word_break_characters = ?
 	//rl_completer_word_break_characters = ?
     	//rl_completion_display_matches_hook = ?
-	rl_attempted_completion_function = NULL;
+          //rl_attempted_completion_function = NULL;
 	//rl_char_is_quoted_p = NULL;
-    	rl_pre_input_hook = NULL;
+    	  //rl_pre_input_hook = NULL;
 
 	warn_writer = default_warn_writer;
         screen_state = SCREEN_UNINITIALIZED;
@@ -455,6 +385,189 @@ screen_finish(void)
     }
 }
 
+/* XXX: move somewhere else. */
+static char *
+dequote_string(const char *str)
+{
+    return dequote_words_full(str, false, false, true, false/*true*/, NULL);
+}
+
+char *
+filename_quote_string(const char *str, bool dquotes, bool finalize)
+{
+    return quote_word_full(str, dquotes, finalize, ";*?", "#", false, true, false/*true*/, true);
+}
+
+char *
+quote_string(const char *str, bool dquotes, bool finalize)
+{
+    return quote_word_full(str, dquotes, finalize, ";", "#", false, true, false/*true*/, true);
+}
+
+void
+fill_completion_info(DCCompletionInfo *ci)
+{
+    ci->word_full = xstrndup(ci->line + ci->ws, ci->we - ci->ws);
+    ci->word = dequote_string(ci->word_full);
+    ci->word_index = get_word_index(ci->line, ci->ws);
+}
+
+#if 0
+/* Strip a trailing unescaped double quote from STR, if there is any.
+ * You can say that this "opens up" STR.
+ */
+static void
+strip_trailing_quote(char *str)
+{
+    bool quoted = false;
+
+    for (; str[0] != '\0'; str++) {
+        if (str[0] == '"') {
+            if (str[1] == '\0' && quoted)
+                str[0] = '\0';
+            quoted = !quoted;
+        } else if (str[0] == '\\' && str[1] != '\0') {
+            str++;
+        }
+    }
+}
+#endif
+
+/* Return escaped text to display for the specified completion entry.
+ */
+static char *
+get_escaped_display(DCCompletionEntry *ce)
+{
+    char *d1;
+    char *d2;
+
+    d1 = xasprintf(ce->display_fmt, ce->display);
+    d2 = quote_word_full(d1, false, false, ";", "#", false, true, false/*true*/, false);
+    free(d1);
+
+    return d2;
+}
+
+/* Return quoted text to put in the command line buffer for the completion entry CE.
+ * If SINGLE is true, then the input_single_fmt may be used.
+ * If FINALIZE is true, then the text will have a trailing closing quote (if quoted)
+ * and a space after that. The returned value should be freed when no longer needed.
+ */
+static char *
+get_quoted_input(DCCompletionInfo *ci, DCCompletionEntry *ce, bool single, bool finalize)
+{
+    char *str1;
+    char *str2;
+
+    /* If quoted already, get rid of trailing quotes (at least temporarily). */
+    if (ce->quoted && ce->input[0] == '"')
+        ce->input[strlen(ce->input)-1] = '\0';
+
+    /* Pick the right input format, and print using it. */
+    if (single && ce->input_single_fmt != NULL) {
+        str1 = xasprintf(ce->input_single_fmt, ce->input);
+    } else {
+        str1 = xasprintf(ce->input_fmt, ce->input);
+    }
+
+    /* Add quotes if necessarily. If already quoted, add trailing quote if it should be there. */
+    if (ce->quoted) {
+        if (ce->input[0] == '"' && ce->finalize && finalize) {
+            str2 = xasprintf("%s\"", str1);
+            free(str1);
+        } else {
+            str2 = str1;
+        }
+    } else {
+        str2 = quote_word_full(str1, ci->line[ci->ws] == '"', finalize && ce->finalize, ";", "#", false, true, true, true);
+        free(str1);
+    }
+
+    if (finalize && ce->finalize) {
+        str1 = xasprintf("%s ", str2);
+        free(str2);
+        return str1;
+    }
+    return str2;
+}
+
+static size_t
+leading_same(const char *c1, const char *c2)
+{
+    int c;
+    for (c = 0; c1[c] != '\0' && c2[c] != '\0'; c++) {
+        if (c1[c] != c2[c])
+            break;
+    }
+    return c;
+}
+
+static int
+completion_readline(int key, int count)
+{
+    char *input;
+    DCCompletionInfo ci;
+    int c;
+
+    ci.line = rl_line_buffer;/*RL*/
+    ci.we = rl_point;/*RL*/
+    ci.ws = find_completion_word_start(ci.line, ci.we);
+    ci.results = ptrv_new();
+    fill_completion_info(&ci);
+
+    default_completion_selector(&ci);
+    if (ci.results->cur == 0) {
+        rl_ding();/*RL*/
+        return -1;
+    }
+
+    if (ci.results->cur == 1) {
+        input = get_quoted_input(&ci, ci.results->buf[0], true, true);
+    } else {
+        int minlen;
+        int maxlen;
+        char *matches[ci.results->cur + 1];
+        DCCompletionEntry *ce;
+
+        ce = ci.results->buf[0];
+        input = get_quoted_input(&ci, ce, false, false);
+        minlen = strlen(input);
+        matches[1] = get_escaped_display(ce);
+        maxlen = strlen(matches[1]);
+
+        for (c = 1; c < ci.results->cur; c++) {
+            char *input2;
+
+            ce = ci.results->buf[c];
+            input2 = get_quoted_input(&ci, ce, false, false);
+            minlen = MIN(minlen, leading_same(input, input2));
+            free(input2);
+            matches[c+1] = get_escaped_display(ce);
+            maxlen = MAX(maxlen, strlen(matches[c+1]));
+        }
+
+        rl_display_match_list(matches, ci.results->cur, maxlen);/*RL*/
+        for (c = 0; c < ci.results->cur; c++)
+            free(matches[c+1]);
+        rl_on_new_line();/*RL*/
+        input[minlen] = '\0';  /* XXX: get rid of trailing escape on input somehow! */
+    }
+
+    rl_begin_undo_group();/*RL*/
+    if (ci.ws != ci.we)
+        rl_point/*RL*/ -= rl_delete_text(ci.ws, ci.we);/*RL*/
+    rl_insert_text(input);/*RL*/ /* automaticly updates rl_point */
+    rl_end_undo_group();/*RL*/
+    free(input);
+
+    free(ci.word_full);
+    free(ci.word);
+    ptrv_foreach(ci.results, (PtrVForeachCallback) free_completion_entry);
+    ptrv_free(ci.results);
+
+    return 0;
+}
+
 /* Prepare the screen prior to waiting for events with select/poll/epoll.
  * Redisplay the prompt if it was cleared by a call to screen_(v)put(f).
  */
@@ -472,15 +585,16 @@ screen_prepare(void)
 	screen_prompt = xasprintf("%s> ", PACKAGE);
 
 	rl_readline_name = PACKAGE;
-	rl_attempted_completion_function = attempted_completion;
-	rl_completer_quote_characters = "\"";
+	rl_add_defun(PACKAGE "-complete", completion_readline, '\t');
+	  //rl_attempted_completion_function = attempted_completion;
+	  //rl_completer_quote_characters = "\"";
 	/* rl_basic_word_break_characters = ... */
-	rl_special_prefixes = "\"";
-	rl_completer_word_break_characters = " \t\n;\"";  // also \"
+	  //rl_special_prefixes = "\"";
+ 	  //rl_completer_word_break_characters = " \t\n;\"";  // also \"
 	/* rl_filename_quote_characters = " \t\n\\\"'>;|&()*?[]~!"; */
 	/* rl_filename_quoting_function = quote_argument; */
 	/* rl_filename_dequoting_function = dequote_argument_rl; */
-	rl_char_is_quoted_p = char_is_quoted;
+          //rl_char_is_quoted_p = char_is_quoted;
     	rl_pre_input_hook = fix_winch;
 
         using_history();
