@@ -220,10 +220,10 @@ match_search_patterns(const char *text, DCSearchSelection *data)
 {
     uint32_t c;
     for (c = 0; c < data->patterncount; c++) {
-	if (match_search_pattern(text, data->patterns+c))
-	    return 1;
+	if (!match_search_pattern(text, data->patterns+c))
+	    return 0;
     }
-    return 0;
+    return 1;
 }
 
 static void
