@@ -37,11 +37,13 @@
 #include "xalloc.h"		/* Gnulib */
 #include "minmax.h"		/* Gnulib */
 #include "quotearg.h"		/* Gnulib */
+#include "memmem.h"		/* Gnulib */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
 #define N_(s) gettext_noop(s)
 #include "common/error.h"
 #include "common/intutil.h"
+#include "common/tempfailure.h"
 #include "microdc.h"
 
 #define LOCK_STRING "ABCABCABCABCABCABCABCABCABCABCABCA"
