@@ -1,6 +1,6 @@
 /* util.c - Utility functions that didn't make it elsewhere
  *
- * Copyright (C) 2004-2005 Oskar Liljeblad
+ * Copyright (C) 2004, 2005 Oskar Liljeblad
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -12,9 +12,9 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU Library General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation,
+ * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
 #include <config.h>
@@ -173,7 +173,7 @@ sockaddr_in_str(struct sockaddr_in *addr)
 }
 
 PtrV *
-wordwrap(char *str, size_t len, size_t first_width, size_t other_width)
+wordwrap(const char *str, size_t len, size_t first_width, size_t other_width)
 {
     size_t width;
     PtrV *out;
