@@ -29,6 +29,8 @@
 #include "quotearg.h"		/* Gnulib */
 #include "minmax.h"		/* Gnulib */
 #include "xvasprintf.h"		/* Gnulib */
+#include "strnlen.h"		/* Gnulib */
+#include "xstrndup.h"		/* Gnulib */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
 #define N_(s) gettext_noop(s)
@@ -38,9 +40,24 @@
 #include "common/quoting.h"
 #include "common/comparison.h"
 #include "common/bksearch.h"
-#include "common/quoting.h"
+//#include "common/quoting.h"
 #include "microdc.h"
 
+/* TRANSLATORS: This comma-separated list specifies
+ * all strings which are considered negative boolean
+ * values for variables such as `active'.
+ * Retain the default English strings and add the
+ * localized ones.
+ */
+#define NEGATIVE_BOOL_STRINGS _("0,off,no,false")
+/* TRANSLATORS: This comma-separated list specifies
+ * all strings which are considered positive boolean
+ * values for variables such as `active'.
+ * Retain the default English strings and add the
+ * localized ones.
+ */
+#define POSITIVE_BOOL_STRINGS _("1,on,yes,true")
+
 static void var_set_nick(DCVariable *var, int argc, char **argv);
 static void var_set_password(DCVariable *var, int argc, char **argv);
 static void var_set_description(DCVariable *var, int argc, char **argv);
@@ -62,9 +79,10 @@ static void var_set_listen_port(DCVariable *var, int argc, char **argv);
 static char *var_get_display_flags(DCVariable *var);
 static void var_set_display_flags(DCVariable *var, int argc, char **argv);
 static void var_set_log_file(DCVariable *var, int argc, char **argv);
-static void speed_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
-static void variable_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
-static void display_completion_generator(const char *line,int ws,int we,char *word,PtrV *results);
+static void speed_completion_generator(DCCompletionInfo *ci);
+static void bool_completion_generator(DCCompletionInfo *ci);
+static void variable_completion_generator(DCCompletionInfo *ci);
+static void display_completion_generator(DCCompletionInfo *ci);
 
 typedef struct {
     DCDisplayFlag flag;
@@ -102,7 +120,7 @@ static DCVariable variables[] = {
     {
       "active",
       var_get_bool, var_set_active, &is_active,
-      NULL,
+      bool_completion_generator,
       NULL,
       "Enable if listening for remote connections"
     },
@@ -380,22 +398,64 @@ var_get_slots(DCVariable *var)
     return xstrdup(uint32_str(my_ul_slots));
 }
 
+/* Improve: return string position or word index in csv where value was found
+ * XXX: move to strutil.c or something. Also on gmediaserver!
+ */
+
+bool
+string_in_csv(const char *csv, char sep, const char *value)
+{
+    const char *p0;
+    const char *p1;
+    size_t len;
+
+    len = strlen(value);
+    for (p0 = csv; (p1 = strchr(p0, sep)) != NULL; p0 = p1+1) {
+        if (p1-p0 == len && strncmp(p0, value, len) == 0)
+            return true;                                 
+    }
+    if (strcmp(p0, value) == 0)
+        return true;
+
+    return false;
+}
+
+static void
+bool_completion_generator(DCCompletionInfo *ci)
+{
+    const char *p0;
+    const char *p1;
+    size_t wlen;
+
+    wlen = strlen(ci->word);
+
+    p0 = POSITIVE_BOOL_STRINGS;
+    for (; (p1 = strchr(p0, ',')) != NULL; p0 = p1+1) {
+        if (p1-p0 >= wlen && memcmp(ci->word, p0, wlen) == 0)
+            ptrv_append(ci->results, new_completion_entry(xstrndup(p0, p1-p0), NULL));
+    }
+    if (strleftcmp(ci->word, p0) == 0)
+        ptrv_append(ci->results, new_completion_entry(xstrdup(p0), NULL));
+
+    p0 = NEGATIVE_BOOL_STRINGS;
+    for (; (p1 = strchr(p0, ',')) != NULL; p0 = p1+1) {
+        if (p1-p0 >= wlen && memcmp(ci->word, p0, wlen) == 0)
+            ptrv_append(ci->results, new_completion_entry(xstrndup(p0, p1-p0), NULL));
+    }
+    if (strleftcmp(ci->word, p0) == 0)
+        ptrv_append(ci->results, new_completion_entry(xstrdup(p0), NULL));
+}
+
 static bool
 parse_bool(const char *s, bool *value)
 {
-    if (strcmp(s, "0") == 0
-	    || strcmp(s, _("off")) == 0
-	    || strcmp(s, _("no")) == 0
-	    || strcmp(s, _("true")) == 0) {
-    	*value = false;
+    if (string_in_csv(POSITIVE_BOOL_STRINGS, ',', s)) {
+	*value = true;
 	return true;
     }
-    if (strcmp(s, "1") == 0
-	    || strcmp(s, _("on")) == 0
-	    || strcmp(s, _("yes")) == 0
-	    || strcmp(s, _("false")) == 0) {
-    	*value = true;
-	return true;
+    if (string_in_csv(NEGATIVE_BOOL_STRINGS, ',', s)) {
+        *value = false;
+        return true;
     }
     return false;
 }
@@ -611,32 +671,33 @@ var_set_password(DCVariable *var, int argc, char **argv)
  * non-match is found.
  */
 void
-variable_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
+variable_completion_generator(DCCompletionInfo *ci)
 {
-    sorted_list_completion_generator(word,results,variables,variables_count,sizeof(DCVariable),offsetof(DCVariable, name));
+    sorted_list_completion_generator(ci->word,ci->results,variables,variables_count,sizeof(DCVariable),offsetof(DCVariable, name));
 }
 
 static void
-speed_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
+speed_completion_generator(DCCompletionInfo *ci)
 {
     int c;
 
     /* There's not enough speed types to justify a binary search... */
     for (c = 0; c < speeds_count; c++) {
-	if (strleftcmp(word, speeds[c]) == 0)
-	    ptrv_append(results, new_completion_entry(speeds[c], NULL));
+	if (strleftcmp(ci->word, speeds[c]) == 0)
+	    ptrv_append(ci->results, new_completion_entry(speeds[c], NULL));
     }
 }
 
 static void
-display_completion_generator(const char *line,int ws,int we,char *word,PtrV *results)
+display_completion_generator(DCCompletionInfo *ci)
 {
+    char *word;
     char base_char;
     uint32_t c;
     uint32_t flags;
     char *cmd;
 
-    cmd = get_word_dequoted(line, 1);
+    cmd = get_word_dequoted(ci->line, 1);
     if (strcmp(cmd, "log") == 0) { /* this is safe even when using aliases */
 	flags = log_flags;
     } else {
@@ -644,6 +705,7 @@ display_completion_generator(const char *line,int ws,int we,char *word,PtrV *res
     }
     free(cmd);
 
+    word = ci->word;
     base_char = (word[0] == '+' || word[0] == '-' ? *word++ : '\0');
     /* XXX: complete `all', `default' */
     for (c = 0; c < display_flag_count; c++) {
@@ -658,7 +720,7 @@ display_completion_generator(const char *line,int ws,int we,char *word,PtrV *res
             entry = new_completion_entry(NULL, NULL);
             entry->input = xasprintf("%c%s", base_char, display_flag_details[c].name);
             entry->display = xstrdup(display_flag_details[c].name);
-	    ptrv_append(results, entry);
+	    ptrv_append(ci->results, entry);
         }
     }
 }
@@ -710,19 +772,19 @@ cmd_set(int argc, char **argv)
 }
 
 void
-set_command_completion_selector(const char *line,int ws,int we,char *word,PtrV *results)
+set_command_completion_selector(DCCompletionInfo *ci)
 {
     char *name;
     DCVariable *var;
 
-    /* get_word_index(..) > 0 is assured by get_completor. */
-    if (get_word_index(line, we) <= 1) { 
-	variable_completion_generator(line,ws,we,word,results);
+    /* ci->word_index > 0 is assured by get_completor. */
+    if (ci->word_index <= 1) {
+	variable_completion_generator(ci);
     } else {
-	name = get_word_dequoted(line, 1);
+	name = get_word_dequoted(ci->line, 1);
 	var = find_variable(name);
 	if (var != NULL && var->completor != NULL)
-	    var->completor(line,ws,we,word,results);
+	    var->completor(ci);
 	free(name);
     }
 }
